// SPDX-FileCopyrightText: Matteo Settenvini <matteo.settenvini@montecristosoftware.eu>
// SPDX-License-Identifier: GPL-3.0-or-later

use {
    serde::{Deserialize, Serialize},
    std::net::IpAddr,
    zbus::{dbus_proxy, Result},
};

#[dbus_proxy(
    default_service = "com.endlessm.ParentalControls",
    interface = "com.endlessm.ParentalControls.Dns",
    default_path = "/com/endlessm/ParentalControls/Dns",
    gen_blocking = false
)]
trait MalcontentDns {
    fn get_dns(&self) -> Result<Restrictions>;
}

#[derive(Clone, Debug, PartialEq, Eq, Serialize, Deserialize, zvariant::Type)]
pub struct Restriction {
    pub ip: IpAddr,
    pub hostname: String,
}

pub type Restrictions = Vec<Restriction>;

pub async fn restrictions() -> anyhow::Result<Vec<Restriction>, anyhow::Error> {
    #[cfg(not(feature = "integration_test"))]
    let connection = {
        // This is the normal behavior
        zbus::Connection::session().await?
    };

    #[cfg(feature = "integration_test")]
    let connection = {
        use tokio::net::TcpStream;

        // During integration testing, we want to connect to a private
        // bus name to avoid clashes with existing system services.
        let socketaddr = std::env::var("TEST_DBUS_SOCKET")
            .expect("The test has not set the TEST_DBUS_SOCKET environment variable");

        let socket = loop {
            match TcpStream::connect(&socketaddr).await {
                Ok(stream) => break stream,
                Err(e) if e.kind() == std::io::ErrorKind::ConnectionRefused => {
                    tokio::task::yield_now().await;
                    continue;
                }
                Err(e) => anyhow::bail!(e),
            }
        };

        log::trace!("connecting to DBus on {}", socket.local_addr()?);
        zbus::ConnectionBuilder::tcp_stream(socket)
            .p2p()
            .build()
            .await?
    };

    let proxy = MalcontentDnsProxy::builder(&connection).build().await?;
    let restrictions = proxy.get_dns().await;
    log::trace!("malcontent-nss: user restrictions are {:?}", &restrictions);
    Ok(restrictions?)
}
